import 'dart:convert';

import 'package:flutter/widgets.dart';
import 'package:http/http.dart' as http;
import 'package:logging/logging.dart';

import '../models/runtime_setting.dart';
import '../models/seed.dart';

class SeedResponse {
  final Seed? seed;
  final UserErrorType? error;

  SeedResponse(this.seed, {this.error});

  bool get hasError => error != null;
}

class UserApi {
  final Logger _log = Logger('UserApi');

  final String _baseUrl;
  final http.Client _client;

  UserApi(this._client, this._baseUrl);

  http.Client get client => _client;

  Future<Color?> getCurrentColor() async {
    final response = await _client.get(Uri.parse('$_baseUrl/current_color'));

    if (response.statusCode == 200) {
      return Color(int.parse(json.decode(response.body)['color'] as String, radix: 16));
    }
  }

  Future<Map<String, dynamic>> getSettings() async {
    final response = await _client.get(Uri.parse('$_baseUrl/settings'));

    if (response.statusCode == 200 && response.body.isNotEmpty) {
      final Map<String, dynamic> settings = json.decode(response.body);
      for (var setting in settings.keys) {
        if (setting.contains('runtime')) {
          settings[setting] = RuntimeSetting.fromMap(json.decode(settings[setting]));
        }
      }
      return settings;
    } else {
      return {};
    }
  }

  Future<SeedResponse> makeSeed(String ticket) async {
    _log.fine('makeSeed for ticket "$ticket"');
    final response = await _client.post(
      Uri.parse('$_baseUrl/makeseed'),
      body: json.encode(ticket),
      headers: {'Content-Type': 'application/json'},
    );

    if (response.statusCode == 200) {
      return SeedResponse(Seed.fromString(response.body));
    } else if (response.statusCode >= 500) {
      _log.severe('server error (${response.statusCode}): ${response.body}');
      return SeedResponse(null, error: UserErrorType.server);
    } else if (response.statusCode >= 400) {
      _log.severe('bad response (${response.statusCode}): ${response.body}');
      if (response.body.contains('encryption')) {
        return SeedResponse(null, error: UserErrorType.encryption);
      } else if (response.body.contains('expired')) {
        return SeedResponse(null, error: UserErrorType.expired);
      } else {
        return SeedResponse(null, error: UserErrorType.other);
      }
    }
    return SeedResponse(null, error: UserErrorType.other);
  }
}

enum UserErrorType {
  signing,
  encryption,
  expired,
  server,
  other,
}
