import 'package:digital_3g_common/backend_switcher.dart';
import 'package:digital_3g_common/theme.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:provider/provider.dart';

import '../api/admin_api.dart';
import '../api/auth_api.dart';
import '../models/user.dart';
import '../widgets/card_list_tile.dart';
import '../widgets/code_dialog.dart';
import '../widgets/delete_dialog.dart';
import '../widgets/user_dialog.dart';

enum CreateAction {
  single,
  codes,
}

class UsersPage extends StatefulWidget {
  const UsersPage({Key? key}) : super(key: key);

  @override
  _UsersPageState createState() => _UsersPageState();
}

class _UsersPageState extends State<UsersPage> {
  AdminApi? _adminApi;

  List<User> _users = [];

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: const Text('Nutzer'),
        actions: [
          IconButton(
            onPressed: () => Navigator.of(context).pushNamed('/admin/users/codes'),
            icon: const Icon(Icons.list_alt_rounded),
            tooltip: 'Zugangscodes anzeigen',
          ),
          PopupMenuButton<CreateAction>(
            onSelected: _onCreate,
            itemBuilder: (BuildContext context) => [
              PopupMenuItem<CreateAction>(
                value: CreateAction.codes,
                child: Row(
                  mainAxisAlignment: MainAxisAlignment.spaceBetween,
                  children: const [
                    Text('Zugangscodes erstellen'),
                    Icon(Icons.post_add_rounded, color: darkGrey)
                  ],
                ),
              ),
              PopupMenuItem<CreateAction>(
                value: CreateAction.single,
                child: Row(
                  mainAxisAlignment: MainAxisAlignment.spaceBetween,
                  children: const [
                    Text('Einzelnen Nutzer erstellen'),
                    Icon(Icons.person_add_alt_1_rounded, color: darkGrey)
                  ],
                ),
              ),
            ],
          ),
        ],
      ),
      body: Center(
        child: ConstrainedBox(
          constraints: const BoxConstraints(maxWidth: 640.0),
          child: RefreshIndicator(
            onRefresh: _onUpdateUsers,
            child: ListView(
              padding: const EdgeInsets.all(8.0),
              children: [
                ..._users.map<Widget>(
                  (User user) => CardListTile(
                    leading: Icon(user.scope == 'admin' ? Icons.verified_user : Icons.person),
                    title: Text(user.fullName),
                    onLongPress: () => showDialog(
                      context: context,
                      builder: (BuildContext context) => SimpleDialog(
                        children: [
                          ListTile(
                            enabled: user.username != Provider.of<AuthApi>(context).user?.username,
                            title: const Text('Löschen'),
                            onTap: () async {
                              Navigator.of(context).pop();
                              await _onDeleteUser(user);
                            },
                          ),
                        ],
                      ),
                    ),
                    actions: [
                      IconButton(
                        onPressed: () => _onEditUser(user: user),
                        icon: const Icon(Icons.edit),
                      ),
                    ],
                  ),
                ),
              ],
            ),
          ),
        ),
      ),
    );
  }

  @override
  void initState() {
    super.initState();
    _initApi();
  }

  void _initApi() async {
    final _client = Provider.of<AuthApi>(context, listen: false).client;
    _adminApi = AdminApi(_client, BlocProvider.of<BackendSwitcher>(context).state);

    _onUpdateUsers();
    super.didChangeDependencies();
  }

  void _onCreate(CreateAction action) {
    if (action == CreateAction.single) {
      _onEditUser();
    } else if (action == CreateAction.codes) {
      showDialog(context: context, builder: (BuildContext context) => const CodeDialog());
    }
  }

  Future<void> _onDeleteUser(User user) async {
    final result = await showDialog(
      context: context,
      builder: (BuildContext context) => DeleteDialog(name: user.username),
    );

    if (result) {
      await _adminApi!.deleteUser(user.id!);
      setState(() {
        _users.removeWhere((element) => element.id == user.id);
      });
    }
  }

  void _onEditUser({User? user}) async {
    final result = await showDialog<UserResult>(
      context: context,
      builder: (BuildContext context) => UserDialog(user: user),
    );

    if (!(result?.save ?? false) || (result?.user?.username.isEmpty ?? false)) return;

    if (user != null) {
      setState(() {
        user = result!.user!;
      });
      await _adminApi!.updateUser(user!);
    } else {
      _users.add((await _adminApi!.addUser(result!.user!, result.password!))!);
      setState(() {
        _users;
      });
    }
  }

  Future<void> _onUpdateUsers() async {
    _users = (await _adminApi!.getUsers()) ?? [];
    setState(() => _users);
  }
}
