import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:connectivity/connectivity.dart';
import 'package:digital_3g_common/backend_switcher.dart';
import 'package:digital_3g_common/content_dialog.dart';
import 'package:digital_3g_common/theme.dart';
import 'package:digital_3g_common/ui.dart';
import 'package:digital_3g_common/verification_api.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:intl/intl.dart';
import 'package:qr_code_scanner/qr_code_scanner.dart';
import 'package:shared_preferences/shared_preferences.dart';

import 'config.dart';
import 'error_column.dart';
import 'verification_bloc.dart';

enum Action {
  info,
  revoke,
}

class HomePage extends StatefulWidget {
  const HomePage({Key? key}) : super(key: key);

  @override
  State<HomePage> createState() => _HomePageState();
}

class _HomePageState extends State<HomePage> {
  final GlobalKey _qrKey = GlobalKey(debugLabel: 'QR');

  final DateFormat _dateFormat = DateFormat.yMMMMd();
  final DateFormat _timeFormat = DateFormat.Hms();

  VerificationBloc? _bloc;

  DateTime _currentDate = DateTime.now();

  Timer? _clockTimer;
  Timer? _refreshTimer;

  QRViewController? _qrController;
  Barcode? _qrResult;
  String _lastResult = '';

  StreamSubscription<ConnectivityResult>? _connectionListener;
  bool _disconnected = false;

  bool _showHelp = false;

  @override
  Widget build(BuildContext context) {
    return Scaffold(
      appBar: AppBar(
        title: Image.asset(
          'assets/uni_logo.png',
          package: 'digital_3g_common',
          height: 40.0,
        ),
        actions: [
          IconButton(
            onPressed: () async {
              final connectivity = await Connectivity().checkConnectivity();
              _throwConnectivityResult(connectivity);
              try {
                _bloc!.onCheck();
              } catch (e) {
                ScaffoldMessenger.of(context).showSnackBar(const SnackBar(
                  content: Text('Nachweis konnte nicht überprüft werden, bitte starte die App neu'),
                ));
              }
            },
            icon: const Icon(Icons.refresh_rounded),
            tooltip: 'Aktualisieren',
          ),
          IconButton(
            onPressed: () => Navigator.of(context).pushNamed('/settings'),
            icon: const Icon(Icons.settings_rounded),
            tooltip: 'Einstellungen',
          ),
        ],
      ),
      floatingActionButton: _disconnected
          ? FloatingActionButton(
              onPressed: () => _bloc!.onCheck(),
              backgroundColor: Colors.red,
              tooltip: 'Internetverbindung unterbrochen',
              child: const Icon(Icons.wifi_off_rounded, color: Colors.white),
            )
          : null,
      body: BlocBuilder<VerificationBloc, VerificationState>(
        bloc: _bloc,
        builder: (BuildContext context, VerificationState state) {
          return Center(
            child: state == VerificationState.busy
                ? const CircularProgressIndicator()
                : ConstrainedBox(
                    constraints: const BoxConstraints(maxWidth: 512.0),
                    child: Column(
                      mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                      crossAxisAlignment: CrossAxisAlignment.stretch,
                      children: [
                        PaddedCard(
                          child: Column(
                            crossAxisAlignment: CrossAxisAlignment.start,
                            mainAxisAlignment: MainAxisAlignment.start,
                            children: [
                              Text(
                                _timeFormat.format(_currentDate.toLocal()),
                                style: Theme.of(context).textTheme.headline5,
                                textScaleFactor: 1.5,
                              ),
                              Text(
                                _dateFormat.format(_currentDate.toLocal()),
                                style: Theme.of(context).textTheme.headline5,
                              ),
                            ],
                          ),
                        ),
                        if (state != VerificationState.noTicket)
                          PaddedCard(
                            child: Column(
                              children: [
                                if (state == VerificationState.verified)
                                  if (_bloc!.verification?.color != null)
                                    Text(
                                      '3G-Nachweis vorhanden',
                                      style: Theme.of(context)
                                          .textTheme
                                          .headline4!
                                          .copyWith(color: const Color(0xFF33ce55)),
                                      textAlign: TextAlign.center,
                                    )
                                  else
                                    ErrorColumn(
                                      'Heute ist keine Farbe vorhanden',
                                      actions: [
                                        ElevatedButton(
                                          onPressed: () => showDialog(
                                              context: context,
                                              builder: (_) => const ContentDialog(
                                                    title: Text('Hilfe'),
                                                    children: [
                                                      Text(
                                                          'Für den heutigen Tag wurde vom Administrator keine Bändchenfarbe festgelegt.\nAn Sonntagen ist das ganz normal, weil die Uni dann nicht geöffnet hat. Bitte nicht den Nachweis löschen!'),
                                                    ],
                                                  )),
                                          child: const Text('Warum?'),
                                        ),
                                      ],
                                    )
                                else if (state == VerificationState.unverified)
                                  ErrorColumn(
                                    'Dein Nachweis ist nicht mehr gültig',
                                    actions: [
                                      ElevatedButton(
                                        onPressed: _onDeleteTicket,
                                        child: const Text('Nachweis löschen'),
                                      ),
                                      ElevatedButton(
                                        onPressed: () => showDialog(
                                            context: context,
                                            builder: (_) => const ContentDialog(
                                                  title: Text('Hilfe'),
                                                  children: [
                                                    Text(
                                                        'Der Endzeitpunkt deines Nachweises wurde erreicht. Das bedeutet, dass du deinen aktuellen Nachweis nicht weiter nutzen kannst. Du solltest deinen Nachweis löschen und an der Akkreditierungsstelle einen neuen Nachweis einholen.'),
                                                    Divider(),
                                                    Text(
                                                        'Die verschiedenen Nachweise haben ab dem Scannen des QR-Codes folgende Laufzeiten\n'),
                                                    Text('Test: 24 Stunden\n'),
                                                    Text(
                                                        'Genesen: Ein halbes Jahr, maximal bis zum Semesterende\n'),
                                                    Text('Geimpft: Bis zum Semesterende\n'),
                                                  ],
                                                )),
                                        child: const Text('Was bedeutet das?'),
                                      ),
                                    ],
                                  )
                                else if (state == VerificationState.error)
                                  if (_bloc!.error == VerificationErrorType.encryption)
                                    ErrorColumn(
                                      'Der Schlüssel deines Nachweises ist ungültig',
                                      actions: [
                                        ElevatedButton(
                                          onPressed: _onDeleteTicket,
                                          child: const Text('Nachweis löschen'),
                                        ),
                                        ElevatedButton(
                                          onPressed: () => showDialog(
                                              context: context,
                                              builder: (_) => const ContentDialog(
                                                    title: Text('Hilfe'),
                                                    children: [
                                                      Text(
                                                          'Der Schlüssel, mit dem alle Nachweise verschlüsselt wurden, wurde ausgetauscht.\nDas macht alle Nachweise, die seit davor im Umlauf sind ungültig.\n\nUm das Problem zu lösen lösche bitte deinen Nachweis und hole dir an der Akkreditierungsstelle einen neuen Nachweis.'),
                                                    ],
                                                  )),
                                          child: const Text('Was bedeutet das?'),
                                        ),
                                      ],
                                    )
                                  else if (_bloc!.error == VerificationErrorType.oldseed)
                                    ErrorColumn(
                                      'Dieser QR-Code ist bereits abgelaufen',
                                      actions: [
                                        ElevatedButton(
                                          onPressed: _bloc!.onRevoke,
                                          child: const Text('Erneut probieren'),
                                        ),
                                        ElevatedButton(
                                          onPressed: () => showDialog(
                                              context: context,
                                              builder: (_) => const ContentDialog(
                                                    title: Text('Hilfe'),
                                                    children: [
                                                      Text(
                                                          'Als weiteres Sicherheitsfeature sind die QR-Codes vom Wachpersonal nur für sehr kurze Zeit nach der Erstellung gültig.\n\nBitte frage das Personal nach einem neuen QR-Code und probiere es noch mal.'),
                                                    ],
                                                  )),
                                          child: const Text('Was bedeutet das?'),
                                        ),
                                      ],
                                    )
                                  else if (_bloc!.error == VerificationErrorType.signing)
                                    ErrorColumn(
                                      'Die Signatur deines Nachweises ist ungültig',
                                      actions: [
                                        BlocProvider.of<BackendSwitcher>(context).mode ==
                                                BackendMode.development
                                            ? ElevatedButton(
                                                onPressed: () =>
                                                    Navigator.of(context).pushNamed('/settings'),
                                                child: const Text('Zu den Einstellungen'),
                                              )
                                            : ElevatedButton(
                                                onPressed: _onDeleteTicket,
                                                child: const Text('Nachweis löschen'),
                                              ),
                                        ElevatedButton(
                                          onPressed: () => showDialog(
                                              context: context,
                                              builder: (_) => const ContentDialog(
                                                    title: Text('Hilfe'),
                                                    children: [
                                                      Text(
                                                          'Jeder Nachweis wurde vom Server signiert um die Identität des Erstellers sicherzustellen.\nEs scheint als wäre der Nachweis vom jemand Anderem signiert oder der QR-Code fehlerhaft gelesen worden.\n\n'),
                                                      Text(
                                                          'Bitte stelle sicher, dass in den Einstellungen der Punkt "Debug-Backend nutzen" deaktiviert ist. Ansonsten lösche bitte deinen Nachweis und probiere es mit einem Neuen.')
                                                    ],
                                                  )),
                                          child: const Text('Was bedeutet das?'),
                                        ),
                                      ],
                                    )
                                  else if (_bloc!.error == VerificationErrorType.server)
                                    ErrorColumn(
                                      'Es gibt ein Problem bei der Serververbindung',
                                      actions: [
                                        ElevatedButton(
                                          onPressed: _bloc!.onCheck,
                                          child: const Text('Erneut probieren'),
                                        ),
                                      ],
                                    )
                                  else
                                    Text(
                                      'Ein unbekanntes Problem wurde festgestellt',
                                      style: Theme.of(context).textTheme.headline4,
                                      textAlign: TextAlign.center,
                                    ),
                              ],
                            ),
                          ),
                        BlocBuilder<BackendSwitcher, String>(
                          builder: (BuildContext context, String state) {
                            if (state == debugBaseUrl) {
                              return PaddedCard(
                                child: Text(
                                  'DEBUG VERSION',
                                  textAlign: TextAlign.center,
                                  style: Theme.of(context)
                                      .textTheme
                                      .headline4
                                      ?.copyWith(color: Colors.red),
                                ),
                              );
                            }
                            return const SizedBox(width: 0.0, height: 0.0);
                          },
                        ),
                        if (state != VerificationState.noTicket &&
                            state != VerificationState.error &&
                            state != VerificationState.unverified)
                          Expanded(
                            child: Padding(
                              padding: const EdgeInsets.all(16.0),
                              child: _bloc!.verification?.color != null
                                  ? ColorCircle(
                                      color: state == VerificationState.verified
                                          ? Color(_bloc!.verification!.color ?? Colors.grey.value)
                                          : Colors.grey,
                                      borderSide: BorderSide.none,
                                      inverse: state == VerificationState.verified
                                          ? _bloc!.verification!.inverse ?? false
                                          : false,
                                    )
                                  : Container(),
                            ),
                          )
                        else if (state != VerificationState.noTicket)
                          const Spacer(),
                        if (_showHelp)
                          Card(
                            clipBehavior: Clip.antiAlias,
                            margin: const EdgeInsets.all(8.0),
                            child: Padding(
                              padding: const EdgeInsets.fromLTRB(8.0, 32.0, 8.0, 32.0),
                              child: Column(
                                children: [
                                  Padding(
                                    padding: const EdgeInsets.only(bottom: 16.0),
                                    child: Text(
                                      'Datenschutz und Hilfe',
                                      style: Theme.of(context).textTheme.headline4,
                                      textAlign: TextAlign.start,
                                    ),
                                  ),
                                  ListTile(
                                    title: const Text('Datenschutzinformationen'),
                                    trailing: const Icon(Icons.security_rounded),
                                    onTap: () =>
                                        showWebsite(context, 'https://go.upb.de/datenschutz_3digi'),
                                  ),
                                  ListTile(
                                    title: const Text('Hilfe zur App'),
                                    trailing: const Icon(Icons.help_outline_rounded),
                                    onTap: () =>
                                        showWebsite(context, 'https://go.upb.de/hilfe_3digi'),
                                  ),
                                  Padding(
                                    padding: const EdgeInsets.only(top: 16.0),
                                    child: ElevatedButton(
                                      onPressed: _onDismissHelp,
                                      child: const Text('Weiter zum QR-Code Scanner'),
                                    ),
                                  ),
                                ],
                              ),
                            ),
                          ),
                        if (state == VerificationState.noTicket && !_showHelp)
                          Expanded(
                            child: Card(
                              clipBehavior: Clip.antiAlias,
                              margin: const EdgeInsets.all(8.0),
                              child: ClipRRect(
                                borderRadius: BorderRadius.circular(8.0),
                                child: Stack(
                                  alignment: Alignment.topRight,
                                  children: [
                                    QRView(
                                      key: _qrKey,
                                      onQRViewCreated: _onQRViewCreated,
                                      overlay: QrScannerOverlayShape(
                                        borderColor: Colors.green,
                                        borderRadius: 16.0,
                                        borderWidth: 4.0,
                                      ),
                                    ),
                                  ],
                                ),
                              ),
                            ),
                          ),
                      ],
                    ),
                  ),
          );
        },
      ),
    );
  }

  @override
  void dispose() {
    _clockTimer?.cancel();
    _refreshTimer?.cancel();
    _connectionListener?.cancel();
    super.dispose();
  }

  @override
  void initState() {
    super.initState();
    _bloc = BlocProvider.of<VerificationBloc>(context);
    _clockTimer = Timer.periodic(const Duration(seconds: 1), (timer) {
      setState(() {
        _currentDate = DateTime.now();
      });
    });
    _refreshTimer = Timer.periodic(const Duration(seconds: 30), (timer) {
      _bloc!.onCheck();
    });
    _initConnectivity();
    _initPreferences();
    _loadMotd();
  }

  @override
  void reassemble() {
    super.reassemble();
    if (Platform.isAndroid) {
      _qrController?.pauseCamera();
    } else if (Platform.isIOS) {
      _qrController?.resumeCamera();
    }
  }

  Future<void> _initConnectivity() async {
    final connected = await Connectivity().checkConnectivity();
    setState(() => _disconnected = connected == ConnectivityResult.none);
    if (_disconnected) {
      _throwConnectivityResult(connected);
    }
    _connectionListener = Connectivity().onConnectivityChanged.listen(_throwConnectivityResult);
  }

  Future<void> _initPreferences() async {
    final prefs = await SharedPreferences.getInstance();
    setState(() => _showHelp = prefs.getBool('show_help') ?? true);
  }

  Future<void> _loadMotd() async {
    final message = await _bloc!.api!.getMotd();
    if (message.isEmpty) return;

    ScaffoldMessenger.of(context).showSnackBar(SnackBar(
      content: Column(
        mainAxisSize: MainAxisSize.min,
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          Text(
            'Systemnachricht',
            style: Theme.of(context).textTheme.headline5!.copyWith(color: magenta),
          ),
          Text(message)
        ],
      ),
      duration: const Duration(seconds: 20),
    ));
  }

  void _onDeleteTicket() async {
    final result = await showDialog<bool>(
      context: context,
      builder: (BuildContext context) => AlertDialog(
        title: const Text('Der Nachweis ist danach endgültig gelöscht. Bist du dir sicher?'),
        actions: [
          MaterialButton(
            onPressed: () => Navigator.of(context).pop(true),
            child: const Text('Ja'),
          ),
          ElevatedButton(
            onPressed: () => Navigator.of(context).pop(false),
            child: const Text('Nein'),
          ),
        ],
      ),
    );

    if (result ?? false) _bloc!.onRevoke();
  }

  Future<void> _onDismissHelp() async {
    setState(() {
      _showHelp = false;
    });

    final prefs = await SharedPreferences.getInstance();
    prefs.setBool('show_help', _showHelp);
  }

  void _onQRViewCreated(QRViewController controller) {
    _qrController = controller;
    _lastResult = '';
    controller.scannedDataStream
        .where((scanData) => (scanData.code?.isNotEmpty ?? false) && (_lastResult != scanData.code))
        .listen((scanData) {
      if (!_disconnected) {
        // check for valid base64 or else cancel
        try {
          final Uint8List decoded = base64.decode(scanData.code!);
          if (decoded.isEmpty) return;
        } catch (e) {
          return;
        }
        _qrResult = scanData;
        _lastResult = _qrResult!.code!;
        _bloc!.onAdd(_lastResult);
      }
    });
  }

  void _throwConnectivityResult(ConnectivityResult result) {
    switch (result) {
      case ConnectivityResult.none:
        if (_disconnected == false) {
          setState(() => _disconnected = true);
          ScaffoldMessenger.of(context).clearSnackBars();
          ScaffoldMessenger.of(context).showSnackBar(
            const SnackBar(content: Text('Die Internetverbindung wurde unterbrochen')),
          );
        }
        break;
      default:
        if (_disconnected == true) {
          setState(() => _disconnected = false);
          _bloc!.onCheck();
          ScaffoldMessenger.of(context).clearSnackBars();
          ScaffoldMessenger.of(context).showSnackBar(
            const SnackBar(content: Text('Verbindung zum Internet hergestellt')),
          );
        }
        break;
    }
  }
}
