import 'dart:convert';
import 'dart:io';

import 'package:bloc/bloc.dart';
import 'package:connectivity/connectivity.dart';
import 'package:digital_3g_common/verification_api.dart';
import 'package:http/http.dart' as http;
import 'package:logging/logging.dart';
import 'package:path_provider/path_provider.dart';

enum EventType {
  check,
  revoke,
  add,
}

class VerificationBloc extends Bloc<VerificationEvent, VerificationState> {
  final Logger _log = Logger('VerificationBloc');

  File? _ticketStore;
  String? _ticket;

  File? _verificationStore;

  VerificationApi? _api;
  VerificationResponse? _verification;
  VerificationErrorType? _error;

  VerificationBloc(String url) : super(VerificationState.busy) {
    _api = VerificationApi(http.Client(), url);
    _initVerificationStore().then((_) => _initTicketFile().then((value) => _loadTicket()));
  }

  VerificationApi? get api => _api;
  VerificationErrorType? get error => _error;
  String? get ticket => _ticket;
  VerificationResponse? get verification => _verification;

  @override
  Stream<VerificationState> mapEventToState(VerificationEvent event) async* {
    _log.fine('got event $event');
    switch (event.type) {
      case EventType.check:
        if (_ticket != null && _ticket!.isNotEmpty) {
          VerificationResponse? response;
          try {
            response = await _api!.check(_ticket!);
            _log.finer(response);
          } catch (e) {
            _log.severe('failed to check ticket:\n$e');
          }
          if ((response?.verified ?? false) && !(response?.hasError ?? false)) {
            _verification = response;
            await _verificationStore!.writeAsString(json.encode(_verification!.toMap));
            yield VerificationState.verified;
          } else if (((((await Connectivity().checkConnectivity()) == ConnectivityResult.none) ||
                  ((response?.hasError ?? false) &&
                      response!.error == VerificationErrorType.server)) &&
              (_verificationStore?.existsSync() ?? false))) {
            // get verification from local storage
            final loaded =
                VerificationResponse.fromMap(json.decode(await _verificationStore!.readAsString()));
            if (loaded.timestamp.isAfter(DateTime.now().subtract(const Duration(minutes: 30)))) {
              _verification = loaded;
              yield VerificationState.verified;
            } else {
              yield VerificationState.unverified;
            }
          } else if (!(response?.verified ?? false) && !(response?.hasError ?? false)) {
            yield VerificationState.unverified;
          } else {
            _verification = response;
            _error = _verification!.error;
            yield VerificationState.error;
          }
        }
        break;
      case EventType.revoke:
        _ticketStore!.writeAsString('', flush: true);
        _ticket = null;
        _error = null;
        yield VerificationState.noTicket;
        break;
      case EventType.add:
        // get ticket from backend if longer than 728 (backwards compatibility)
        TicketResponse? response;
        if ((event.seed?.length ?? 0) > 768) {
          try {
            response = await _api!.generateTicket(event.seed!);
          } catch (e) {
            _log.severe('error while generating ticket');
            _error = VerificationErrorType.other;
            yield VerificationState.error;
          }
        } else {
          response = TicketResponse(event.seed);
        }

        if (response!.hasError) {
          _error = response.error;
          yield VerificationState.error;
        }

        // save ticket
        if (!_ticketStore!.existsSync()) {
          await _ticketStore!.create(recursive: true);
        }
        if (response.ticket != null && response.ticket!.isNotEmpty) {
          _ticket = response.ticket;
          await _ticketStore!.writeAsString(_ticket!, flush: true);
          yield VerificationState.unverified;
          add(VerificationEvent.check());
        }
        break;
      default:
        return;
    }
  }

  void onAdd(String ticket) => add(VerificationEvent.add(ticket));

  @override
  void onChange(Change<VerificationState> change) {
    _log.finer('new state: ${change.nextState}');
    super.onChange(change);
  }

  void onCheck() => add(VerificationEvent.check());

  void onRevoke() => add(VerificationEvent.revoke());

  Future<void> _initTicketFile() async {
    _ticketStore = File('${(await getApplicationDocumentsDirectory()).path}/ticket');
    _log.fine('initialized ticket store in ${_ticketStore?.path}');
  }

  Future<void> _initVerificationStore() async {
    _verificationStore = File('${(await getApplicationDocumentsDirectory()).path}/last');
    _log.fine('initialized verification store in ${_verificationStore?.path}');
  }

  Future<void> _loadTicket() async {
    String? ticket;
    if (!_ticketStore!.existsSync()) {
      _log.fine('no ticket found, creating ticket store');
      _ticketStore!.create(recursive: true);
      add(VerificationEvent.revoke());
      return;
    }

    _log.fine('ticketStore found, loading and checking');
    ticket = await _ticketStore!.readAsString();
    _log.finer('ticket: "$ticket"');
    if (ticket.isNotEmpty) {
      _ticket = ticket;
      add(VerificationEvent.check());
    } else {
      _log.fine('ticket empty, going to noTicket state');
      add(VerificationEvent.revoke());
    }
  }
}

class VerificationEvent {
  final EventType type;
  final String? seed;

  VerificationEvent({required this.type, this.seed});

  factory VerificationEvent.add(String ticket) =>
      VerificationEvent(type: EventType.add, seed: ticket);

  factory VerificationEvent.check() => VerificationEvent(type: EventType.check);

  factory VerificationEvent.revoke() => VerificationEvent(type: EventType.revoke);

  Map get toMap => {'type': type, 'seed': seed};

  @override
  String toString() => '[VerificationEvent $toMap]';
}

enum VerificationState {
  noTicket,
  verified,
  unverified,
  error,
  busy,
}
